import java.util.*;

class PacketDeparture implements Event {
    Link link;

    public PacketDeparture(Link l) {
	link = l;
    }

    public void process() {
	link.process_departure();
    }
}

class PacketDelivery implements Event {
    Link link;
    Packet packet;

    public PacketDelivery(Link l, Packet p) {
	link = l;
	packet = p;
    }

    public void process() {
	link.process_delivery(packet);
    }
}

class PacketQueue {
    private Packet [] Q;
    private int head;
    private int tail;

    PacketQueue(int size) {
	Q = new Packet[size];
	head = 0;
	tail = 0;
    }

    private int next(int i) {
	i += 1;
	if (i == Q.length)
	    i = 0;
	return i;
    }

    public boolean empty() {
	return head == tail;
    }

    public boolean full() {
	return next(tail) == head;
    }

    public Packet dequeue() {
	if (head == tail)
	    return null;
	Packet p = Q[head];
	head = next(head);
	return p;
    }

    public Packet first() {
	if (head == tail)
	    return null;
	return Q[head];
    }

    public void enqueue(Packet p) {
	int new_tail = next(tail);
	if (new_tail !=  head) {
	    Q[tail] = p;
	    tail = new_tail;
	} else {
	    System.out.println(Simulation.time() + ": tx queue is full: packet dropped");
	}
    }
}

public class Link {
    private double rate = 1000000; 	// 1Mbytes/second
    private double delay = 0.1; 	// 100 milliseconds

    private PacketQueue tx_queue;
    private Receiver receiver;

    public static final int DEFAULT_TX_QUEUE_SIZE = 100;

    public Link(double r, double d) {
	rate = r;
	delay = d;
	tx_queue = new PacketQueue(DEFAULT_TX_QUEUE_SIZE);
    }

    public Link(double r, double d, int tx_qsize) {
	rate = r;
	delay = d;
	tx_queue = new PacketQueue(tx_qsize);
    }

    void connect(Receiver r) {
	receiver = r;
    }

    void send(Packet p) {
	if (tx_queue.empty())
	    Simulation.schedule_event(p.size()/rate, new PacketDeparture(this));
	tx_queue.enqueue(p);
    }

    void process_departure() {
	Packet p = tx_queue.dequeue();
	Simulation.schedule_event(delay, new PacketDelivery(this, p));
	if (!tx_queue.empty())
	    Simulation.schedule_event(tx_queue.first().size()/rate, new PacketDeparture(this));
    }

    void process_delivery(Packet p) {
	receiver.deliver(p);
    }
}
